import pygtk
pygtk.require('2.0')
import gtk

import ewmh
from hyperswitch.filters import CamelHumpFilter
from hyperswitch.windows import ActiveWindowRepository


class PySwitchController(object):
    def __init__(self):
        self._window = PySwitchWindow()
        self._window.connect("delete_event", gtk.main_quit)
        self._window.set_filter_changed_callback(self._filter_changed)

        self._filter = CamelHumpFilter()

        self._ewmh = ewmh.EWMH()
        self._repo = ActiveWindowRepository(self._ewmh)

        self._window_list = gtk.ListStore(object)
        for win in self._repo.load_windows():
            self._window_list.append([win])

        self._window_list = self._window_list.filter_new()
        self._window_list.set_visible_func(self._filter_function)

        self._window.set_window_list(self._window_list)

    def main(self):
        self._window.show()
        gtk.main()

    def _filter_changed(self, new_filter):
        self._filter.filter_string = new_filter
        self._window_list.refilter()

    def _filter_function(self, model, item):
        win = model.get_value(item, 0)
        return self._filter.is_filter_match(win.title)


class PySwitchWindow(gtk.Window):
    def __init__(self):
        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)

        self.set_position(gtk.WIN_POS_CENTER)
        self.set_size_request(640, 250)
        self.set_border_width(5)

        self.connect('key-press-event', self._key_pressed)

        self._vert_box = gtk.VBox(False, 0)
        self.add(self._vert_box)

        # Entry for filter string        
        self._filter_entry = gtk.Entry()
        self._filter_entry.connect("changed", self._filter_entry_changed, None)
        self._filter_entry.connect("activate", self._launch_current, None)
        self._vert_box.pack_start(self._filter_entry, expand=False)
        self._filter_entry.show()

        # List view for active windows
        self._window_list = gtk.TreeView()
        self._window_list.set_enable_search(False)
        self._window_list.set_headers_visible(False)

        self._title_column = gtk.TreeViewColumn('Title')
        self._window_list.append_column(self._title_column)

        self._title_renderer = gtk.CellRendererText()
        self._title_column.pack_start(self._title_renderer)
        self._title_column.set_cell_data_func(self._title_renderer, self._get_window_title)

        self._window_selection = self._window_list.get_selection()
        self._window_selection.set_mode(gtk.SELECTION_SINGLE)

        self._vert_box.pack_start(self._window_list)
        self._window_list.show()

        self._vert_box.show()

    @staticmethod
    def _get_window_title(column, cell, tree_model, iter, user_data=None):
        win = tree_model.get_value(iter, 0)
        cell.set_property('text', win.title)

    def _filter_entry_changed(self, widget, data=None):
        if self._filter_changed_callback is not None:
            self._filter_changed_callback(widget.get_text())

    def _get_selected(self):
        (model, iter) = self._window_selection.get_selected()

        if iter is None:
            # Default to first item
            iter = model.get_iter_first()

        return iter

    def _launch_current(self, widget, data=None):
        selection = self._get_selected()

        current_win = self._window_list_store.get_value(selection, 0)

        current_win.bring_to_front()
        gtk.main_quit()

    def _select_previous_window(self):
        current_selection = self._get_selected()

        # Convert to path to hack an equivalent for the missing iter_prev function
        current_selection = self._window_list_store.get_path(current_selection)
        prev_window = current_selection
        prev_window = prev_window[0] - 1

        # Select the last item if we are currently at the first item
        if prev_window < 0:
            prev_window = len(self._window_list_store) - 1

        self._window_selection.select_path(prev_window)

    def _select_next_window(self):
        current_selection = self._get_selected()

        next_window = self._window_list_store.iter_next(current_selection)

        if next_window is None:
            next_window = self._window_list_store.get_iter_first()

        self._window_selection.select_iter(next_window)

    def _key_pressed(self, widget, event):
        key = gtk.gdk.keyval_name(event.keyval)

        if key == 'Up':
            self._select_previous_window()
            return True
        elif key == 'Down':
            self._select_next_window()
            return True
        elif key == 'Escape':
            gtk.main_quit()
            return True

    def set_window_list(self, windows):
        """
        Sets the list of windows to display.
        windows should be a TreeModel of Window objects
        """
        self._window_list_store = windows
        self._window_list.set_model(self._window_list_store)

    def set_filter_changed_callback(self, callback):
        self._filter_changed_callback = callback




class ActiveWindowRepository(object):
    _exclusion_list = [None, "Desktop"]

    def __init__(self, wm_interface):
        self._wm_interface = wm_interface

    def load_windows(self):
        clients = self._wm_interface.getClientListStacking()

        windows = []

        for client in clients:
            name = self._wm_interface.getWmName(client)
            if name not in ActiveWindowRepository._exclusion_list:
                windows.append(Window(self._wm_interface, client, name))

        # Reverse the order so that the most recent window is on top
        windows.reverse()
        return windows


class Window(object):
    def __init__(self, wm_interface, wm_id, title):
        self._wm_interface = wm_interface
        self.id = wm_id
        self.title = title

    def bring_to_front(self):
        print "Bring to front: " + self.title
        self._wm_interface.setActiveWindow(self.id)
        self._wm_interface.display.flush()

    def __str__(self):
        return "<Window: " + self.title + ">"



class Word(object):
    def __init__(self, word, case_insensitive=True):
        self.word = word
        self.is_case_insensitive = case_insensitive

    def starts_with(self, char):
        """
        Returns True if the word starts with char, False otherwise.
        """
        return self.char_equal(char, 0)

    def char_equal(self, char, index):
        """
        Returns True if the character at index equals char, False otherwise.
        """
        word_char = self.word[index]
        compare_char = char

        if self.is_case_insensitive:
            word_char = word_char.lower()
            compare_char = compare_char.lower()

        return word_char == compare_char

    def __str__(self):
        return "<Word: '" + reduce(lambda s1, s2: s1 + s2, self.word, '') + "'>"

    def __len__(self):
        return len(self.word)


class CamelHumpFilter(object):
    def __init__(self):
        self._filter_string = ''
        self._processed_filter_string = ''

    @property
    def filter_string(self):
        return self._filter_string

    @filter_string.setter
    def filter_string(self, value):
        self._filter_string = value
        self._processed_filter_string = self.process_filter_string(self._filter_string)

    @staticmethod
    def process_filter_string(filter_string):
        if len(filter_string) == 0:
            return []

        return filter(lambda s: s.isalnum(), filter_string.lower())

    @staticmethod
    def find_index(words, predicate, start_index=0):
        for index, word in enumerate(words):
            if index >= start_index and predicate(word):
                return index

        return None

    def is_filter_match(self, query):
        if len(self._processed_filter_string) == 0:
            return True

        words = self.to_word_list(query)

        if len(words) == 0:
            return False

        first_char = self._processed_filter_string[0]

        first_word_index = self.find_index(words, lambda w: w.starts_with(first_char))

        is_match = False
        while not is_match and first_word_index is not None:
            is_match = self.is_match_from_word(words, first_word_index, 0)
            first_word_index = self.find_index(words, lambda w: w.starts_with(first_char), first_word_index + 1)

        return is_match

    def is_match_from_word(self, words, word_index, start_filter_index=0):
        current_word = words[word_index]

        if len(words) > word_index + 1:
            next_word = words[word_index + 1]
        else:
            next_word = None

        next_word_start_indexes = []

        filter_index = start_filter_index
        current_word_index = 0

        while filter_index < len(self._processed_filter_string):
            current_char = self._processed_filter_string[filter_index]

            if (filter_index > start_filter_index and
                    next_word is not None and
                    next_word.starts_with(current_char)):
                next_word_start_indexes.append(filter_index)

            if (len(current_word) > current_word_index and
                    current_word.char_equal(current_char, current_word_index)):

                filter_index += 1
                current_word_index += 1
            else:
                is_match = False
                while not is_match and len(next_word_start_indexes) > 0:
                    is_match = self.is_match_from_word(words, word_index + 1, next_word_start_indexes.pop())

                return is_match

        return True

    @staticmethod
    def to_word_list(query):
        word_list = []
        current_word = []

        if len(query) == 0:
            return word_list

        for char in query:
            if not char.isalnum():
                # new word
                if len(current_word) > 0:
                    word_list.append(Word(current_word))
                    current_word = []

                continue

            if char.isupper() or char.isdigit():
                # New word
                if len(current_word) > 0:
                    word_list.append(Word(current_word))
                    current_word = []

            current_word.append(char)

        if len(current_word) > 0:
            word_list.append(Word(current_word))

        return word_list



from setuptools import setup

setup(
    name='hyperswitch',
    version='0.1',
    description='A context switching application for Linux',
    url='https://bitbucket.org/TheNoctambulist/hyperswitch/',
    author='TheNoctambulist',
    author_email='TheNoctambulist@zoho.com',
    license='MIT License',
    zip_safe=False,
    packages=['hyperswitch'],
    package_data={
        '': ['*.txt', '*.rst']
    },
    entry_points={
        'gui_scripts': [
            'hyperswitch = hyperswitch:main'
        ]
    },
    install_requires=[
        "ewmh"
    ]
)
